(* ===================================================================

  Lingu (= little language) is a small scripting language to script
  transactions on a database. It is defined as an extension to L0.

  This is an embedding of Lingu in HOL. The embedding of L0 is defined
  else where (see L0Script).

  author: wishnu prasetya

=================================================================== *)

local

(*
load "pred_setLib" ;
load "L0Theory" ;
*)

open Parse HolKernel boolLib bossLib 
open L0Theory 
open pred_setLib 

in 

val _ = new_theory "lingu" ;


(* ===================================================================

SYNTAX

Lingu has the mainly same syntax as L0, with the following differences:

--- Instruction

   * while-loop is not allowed
   * additionally Lingu has the following db transactional instructions:

       ins SimpleExpr TableExpr

     | insert TableExpr TableExpr (map Var. SimpleExpr) (only Var. Expr)

     | del SimpleExpr TableExpr

     | delete TableExpr (drop Var. Expr)

     | update TableExpr (map Var. SimpleExpr) (only Var. Expr)


     Additionally, all target tables in the above instructions should be identifier.

--- Expression and Assertion

Both have the same syntax. An expression can be a simple expression
(expression over data) or a table expression:


        Expr --> SimpleExpr | TableExpr

        TableExpr --> select Expr (map Var. SimpleExpr) (only Var. SimpleExpr)

                    | SimpleExpr IN TableExpr
                     
                    | ALLof  TableExpr (satisfy Var. SimpleExpr)

                    | SOMEof TableExpr (satisfy Var. SimpleExpr)


        SimpleExpr --> Identifier 
                     | Constant
                     | UnaryOp SimpleExpr
                     | SimpleExpr BinOp SimpleExpr
                     | SimpleExpr . FieldName         // record selection
                     | <| FieldName := SimpleExpr
                        , ...
                        , FieldName := SimpleExpr
                        |>                            // record forming

        UnaryOp --> ~

        BinaryOp --> numeric ops, boolean ops, = 
             
=================================================================== *)

(* ===================================================================

   Abstractly and semantically, a table t is represented by a set 
   of rows. We define a number of functions deleteX, updateX, and 
   select. They work on tables, their meaning is as given below.

   These functions are not intended to be executed as is. They 
   specify abstract semantic notions.

=================================================================== *)

val deleteX_def = Define `deleteX t p   = { r | r IN t /\ ~p r}` ;
val updateX_def = Define `updateX t f p = {if p r then f r else r | r IN t}` ;
val select_def  = Define `select t f p  = {f r | r IN t /\ p r}` ;
val insertX_def = Define `insertX td t f p
                          = 
                          {f r | r IN td /\ p r} UNION t` ;

(* ===================================================================

   map, only, drop, satisfy phrases are actually alias of lambda-binder

=================================================================== *)

val _ = new_binder_definition 
   ("map_def",
  --`(map:('a->'b)->('a->'b)) body = body`--) ;

val _ = new_binder_definition 
   ("only_def",
  --`(only:('a->bool)->('a->bool)) body = body`--) ;

val _ = new_binder_definition 
   ("drop_def",
  --`(drop:('a->bool)->('a->bool)) body = body`--) ;

val _ = new_binder_definition 
   ("satisfy_def",
  --`(satisfy:('a->bool)->('a->bool)) body = body`--) ;


(* ===================================================================

   We define now three db instructions: insert, delete, and
   update. They are defined here in terms of their abstract 
   semantics.

=================================================================== *)


val ins_def    = Define `ins r t         = (t /:= r INSERT t)` ;
val insert_def = Define `insert td t f p = (t /:= insertX td t f p)` ;
val del_def    = Define `del r t         = (t /:= deleteX t (drop r'. r'=r))` ;
val delete_def = Define `delete t p      = (t /:= deleteX t p)` ;
val update_def = Define `update t f p    = (t /:= updateX t f p)` ;

(* ===================================================================

   ALL and SOME expressions

=================================================================== *)


val ALLof_def  = Define `ALLof  t p = !r. r IN t ==> p r` ; 
val SOMEof_def = Define `SOMEof t p = ?r. r IN t /\ p r` ;


(* ===================================================================

EXAMPLES

Examples of lingu instructions and expressions you can now write:

val example1 = --`insert t1 t2
                     (map r . r+1)
                     (only r. r>0)`-- ;

val example2 = --`delete t (drop r. r=0)`-- ;

val example3 = --`update t
                     (map r . r+1)
                     (only r. r>0)`-- ;

val example4 = --`ALLof (select t (map r. r+1) (only r. r>0))
                        (satisfy r. r>1)`-- ;


An example of a lingu script we can now write in HOL:

val _ = Hol_datatype 
        `Entry
         = 
         <| Nr     : string ; 
	    Desc   : string ;
	    Date   : string ;
	    Debit  : int ;
	    Credit : int 
	|>` ;

val _ = Hol_datatype
        `Transaction
         =
        <|  Nr     : string ; 
	    Desc   : string ;
	    Date   : string ;
	    Amount : int 
	|>` ;


val credit_def = Define 

      `credit (account, trans, REF t)
          =
          pre T
          post ~(t IN trans)
          do
          /{ ins t trans 
           ; ins <| Nr     := t.Nr     ; 
                    Desc   := t.Desc   ; 
                    Date   := t.Date   ;
                    Debit  := 0        ; 
                    Credit := t.Amount |>
                 account 
          ; delete trans (drop r. r.Nr = t.Nr)
         /}
         return void
         ` ;

And HOL will syntax-check and type-check it for you :)

=================================================================== *)

(* ===================================================================
   Here follows a bunch theorems which will be used to normalize
   a specification, as well as the result of a WP computation.
=================================================================== *)

val set_ss = std_ss ++ PRED_SET_ss ;

val lingu_defs = map definition ["map_def", "only_def", "drop_def", "satisfy_def"]
                 @
                 [ deleteX_def, updateX_def, select_def, insertX_def
                 , ins_def, insert_def, del_def, delete_def, update_def
                 , ALLof_def, SOMEof_def ] ;

(* the only  possible predicates about tables are of the form:

      ALLof t p
      SOMEof t p
      r IN t

other predicates are value-level predicates *)

val ALLof_ins_thm = store_thm ("ALLof_ins_thm",
    --`ALLof (r INSERT t) p  =  p r /\ ALLof t p`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val SOMEof_ins_thm = store_thm ("SOMEof_ins_thm",
    --`SOMEof (r INSERT t) p  =  p r \/ SOMEof t p`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val IN_ins_thm = store_thm ("IN_ins_thm",
    --`s IN (r INSERT t) = (s=r) \/ s IN t`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val ALLof_insert_thm = store_thm ("ALLof_insert_thm",
    --`ALLof (insertX td t f p) q  =  ALLof td (\r. p r ==> q (f r)) /\ ALLof t q`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val SOMEof_insert_thm = store_thm ("SOMEof_insert_thm",
    --`SOMEof (insertX td t f p) q  =  SOMEof td (\r. p r /\ q (f r)) \/ SOMEof t q`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val IN_insert_thm = store_thm ("IN_insert_thm",
    --`r IN (insertX td t f p)  =  (?r'. p r' /\ r' IN td /\ (r = f r')) \/ r IN t`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val ALLof_delete_thm = store_thm ("ALLof_delete_thm",
    --`ALLof (deleteX t p) q  =  ALLof t (\r. ~p r ==> q r)`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val SOMEof_delete_thm = store_thm ("SOMEof_delete_thm",
    --`SOMEof (deleteX t p) q  =  SOMEof t (\r. ~p r /\ q r)`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val IN_delete_thm = store_thm ("IN_delete_thm",
    --`r IN deleteX t p  =  ~p r /\ r IN t`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val ALLof_update_thm = store_thm ("ALLof_update_thm",
    --`ALLof (updateX t f p) q  
       =  
       ALLof t (\r. (p r ==> q (f r)) /\ (~p r ==> q r))`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val SOMEof_update_thm = store_thm ("SOMEof_update_thm",
    --`SOMEof (updateX t f p) q  
       =  
       SOMEof t (\r. (p r /\ q (f r)) \/ (~p r /\ q r))`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val IN_update_thm = store_thm ("IN_update_thm",
    --`r IN updateX t f p
       =
       (?r'. r' IN t /\ (p r' ==> (r = f r')) /\ (~p r' ==> (r=r')))`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val ALLof_select_thm = store_thm ("ALLof_select_thm",
    --`ALLof (select t f p) q  
       =  
       ALLof t (\r. p r ==> q (f r))`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val SOMEof_select_thm = store_thm ("SOMEof_select_thm",
    --`SOMEof (select t f p) q  
       =  
       SOMEof t (\r. p r /\ q (f r))`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;

val IN_select_thm = store_thm ("IN_select_thm",
    --`r IN select t f p
       =  
       ?r'. r' IN t /\ p r' /\ (r = f r')`--,
    RW_TAC set_ss lingu_defs
    THEN PROVE_TAC []) ;


(* ======

and here is the normalizer CONV:

val lingu_normalizer_defs = map definition ["map_def", "only_def", "drop_def", "satisfy_def"]
                            @
                            [ ins_def, insert_def, del_def, delete_def, update_def
                            , ALLof_ins_thm, SOMEof_ins_thm, IN_ins_thm
                            , ALLof_insert_thm, SOMEof_insert_thm, IN_insert_thm
                            , ALLof_delete_thm, SOMEof_delete_thm, IN_delete_thm
                            , ALLof_update_thm, SOMEof_update_thm, IN_update_thm
                            , ALLof_select_thm, SOMEof_select_thm, IN_select_thm
                            ] ;


val lingu_normalizer_CONV =
    REWRITE_CONV lingu_normalizer_defs
    THENC REWRITE_CONV [ALLof_def, SOMEof_def, IN_DEF]
    THENC DEPTH_CONV BETA_CONV ;

===== *)
 

val _ = export_theory();

end ;

